@@ -30,6 +30,9 @@
 #include <gpgme++/key.h>
 #include <gpgme++/data.h>
 #include <gpgme++/importresult.h>
+#include <gpgme++/engineinfo.h>
+#include <gpgme++/defaultassuantransaction.h>
+#include <gpgme++/configuration.h>
 
 #include <windows.h>
 
@@ -41,6 +44,7 @@ GPGRT_LOCK_DEFINE (keycache_lock);
 GPGRT_LOCK_DEFINE (fpr_map_lock);
 GPGRT_LOCK_DEFINE (update_lock);
 GPGRT_LOCK_DEFINE (import_lock);
+GPGRT_LOCK_DEFINE (config_lock);
 static KeyCache* singleton = nullptr;
 
 /** At some point we need to set a limit. There
@@ -306,18 +310,176 @@ do_populate_protocol (GpgME::Protocol proto, bool secret)
   TRETURN;
 }
 
+const std::vector< std::pair<std::string, std::string> >
+gpgagent_transact (std::unique_ptr <GpgME::Context> &ctx,
+                   const char *command)
+{
+  GpgME::Error err;
+  err = ctx->assuanTransact (command);
+  std::unique_ptr<GpgME::AssuanTransaction> t = ctx->takeLastAssuanTransaction();
+  std::unique_ptr<GpgME::DefaultAssuanTransaction> d (dynamic_cast<GpgME::DefaultAssuanTransaction*>(t.release()));
+
+  if (d)
+    {
+      return d->statusLines ();
+    }
+  return std::vector< std::pair<std::string, std::string> > ();
+}
+
+static void
+gpgsm_learn ()
+{
+  TSTART;
+  const GpgME::EngineInfo ei = GpgME::engineInfo (GpgME::CMS);
+
+  if (!ei.fileName ())
+    {
+      STRANGEPOINT;
+      TRETURN;
+    }
+
+  std::vector<std::string> args;
+
+  args.push_back (ei.fileName ());
+  args.push_back ("--learn-card");
+
+  // Spawn the process
+  auto ctx = GpgME::Context::createForEngine (GpgME::SpawnEngine);
+
+  if (!ctx)
+    {
+      STRANGEPOINT;
+      TRETURN;
+    }
+
+  GpgME::Data mystdin, mystdout, mystderr;
+
+  char **cargs = vector_to_cArray (args);
+
+  GpgME::Error err = ctx->spawn (cargs[0], const_cast <const char **> (cargs),
+                                 mystdin, mystdout, mystderr,
+                                 GpgME::Context::SpawnNone);
+  release_cArray (cargs);
+
+  if (err)
+    {
+      log_debug ("%s:%s: gpgsm learn spawn code: %i asString: %s",
+                 SRCNAME, __func__, err.code(), err.asString());
+    }
+  if ((opt.enable_debug & DBG_DATA))
+    {
+      log_data ("stdout:\n'%s'\nstderr:\n%s", mystdout.toString ().c_str (),
+                mystderr.toString ().c_str ());
+    }
+  TRETURN;
+}
+
+static bool already_learned = false;
+
+static void
+do_populate_smartcards (GpgME::Protocol proto)
+{
+  TSTART;
+
+  if (already_learned)
+    {
+      /* gpgsm --learn learns for both protocols. */
+      log_dbg ("Already called gpgsm --learn this run.");
+      TRETURN;
+    }
+
+  GpgME::Error err;
+  auto ctx = GpgME::Context::createForEngine (GpgME::AssuanEngine, &err);
+
+  if (err)
+    {
+      log_dbg ("Failed to create assuan engine. %s",
+               err.asString ());
+      TRETURN;
+    }
+  const auto serials = gpgagent_transact (ctx, "scd serialno");
+  if (serials.empty ())
+    {
+      log_dbg ("No smartcard found.");
+    }
+
+  const auto pairinfo = gpgagent_transact (ctx, "scd learn --keypairinfo");
+  /* If we have a supported card output looks like this:
+     S KEYPAIRINFO 84DC19DC8A563302DA14FF33D42FBBB815170E19 NKS-NKS3.4531 sa
+     S KEYPAIRINFO CD9C93F2C76CCA935BC47114C5B527491BA4896D NKS-NKS3.45B1 e
+     S KEYPAIRINFO 033829E72E9007213FE3E1D9FD3D286DB5D07599 NKS-NKS3.45B2 e
+     S KEYPAIRINFO 0404417F832BE6266585190843B38204431EE26A NKS-SIGG.4531 sae
+  */
+  if (pairinfo.empty ())
+    {
+      log_dbg ("Did not find any smartcards.");
+      TRETURN;
+    }
+
+  auto search_ctx = GpgME::Context::create (proto);
+
+  bool need_to_learn = false;
+  for (const auto &info: pairinfo)
+    {
+      if (info.first != "KEYPAIRINFO")
+        {
+          continue;
+        }
+
+      const auto vec = gpgol_split (info.second, ' ');
+      if (!vec.size ())
+        {
+          log_dbg ("Unexpected keypairinfo line '%s'", info.second.c_str ());
+          continue;
+        }
+
+      if (vec.size () >= 2 && starts_with (vec[1], "OPENPGP") &&
+          proto != GpgME::OpenPGP)
+        {
+          log_dbg ("Skipping OpenPGP key %s", anonstr (vec[0].c_str ()));
+          continue;
+        }
+      const auto keygrip = std::string ("&") + vec[0];
+
+      const auto key = search_ctx->key (keygrip.c_str (), err, true);
+      if (err || key.isNull ())
+        {
+          log_debug ("Key for grip: %s not found. Searching.",
+                     keygrip.c_str ());
+          need_to_learn = true;
+          break;
+        }
+    }
+
+  if (need_to_learn)
+    {
+      /* Note: gpgsm learn also works for openpgp as it learns the keygrips
+         for both. */
+      already_learned = true;
+      gpgsm_learn ();
+    }
+  TRETURN;
+}
+
 static DWORD WINAPI
 do_populate (LPVOID)
 {
   TSTART;
 
+  log_dbg ("Populating config");
+  gpgrt_lock_lock (&config_lock);
+  GpgME::Error err;
+  KeyCache::instance ()->setConfig (GpgME::Configuration::Component::load (err));
+  gpgrt_lock_unlock (&config_lock);
   log_debug ("%s:%s: Populating keycache",
              SRCNAME, __func__);
   do_populate_protocol (GpgME::OpenPGP, false);
+  do_populate_smartcards (GpgME::OpenPGP);
   do_populate_protocol (GpgME::OpenPGP, true);
   if (opt.enable_smime)
     {
       do_populate_protocol (GpgME::CMS, false);
+      do_populate_smartcards (GpgME::CMS);
       do_populate_protocol (GpgME::CMS, true);
     }
   log_debug ("%s:%s: Keycache populated",
@@ -330,9 +492,8 @@ do_populate (LPVOID)
 class KeyCache::Private
 {
 public:
-  Private()
+  Private() : m_use_tofu (false)
   {
-
   }
 
   void setPgpKey(const std::string &mbox, const GpgME::Key &key)
@@ -681,6 +842,10 @@ public:
       }
     for (const auto &recip: recipients)
       {
+        if (recip.empty ())
+          {
+            continue;
+          }
         const auto overrides = getOverrides (recip.c_str (), proto);
 
         if (!overrides.empty())
@@ -1087,6 +1252,8 @@ public:
       TRETURN;
     }
 
+
+
   void populate ()
     {
       TSTART;
@@ -1099,6 +1266,19 @@ public:
       TRETURN;
     }
 
+  const std::vector<GpgME::Configuration::Component> get_cached_config () const
+    {
+      /* Quick hack to ensure that the config is loaded. */
+      gpgrt_lock_lock (&config_lock);
+      gpgrt_lock_unlock (&config_lock);
+      return m_cached_config;
+    }
+
+  void setConfig (const std::vector<GpgME::Configuration::Component> &conf)
+    {
+      m_cached_config = conf;
+    }
+
   std::unordered_map<std::string, GpgME::Key> m_pgp_key_map;
   std::unordered_map<std::string, GpgME::Key> m_smime_key_map;
   std::unordered_map<std::string, GpgME::Key> m_pgp_skey_map;
@@ -1113,6 +1293,8 @@ public:
   std::set<std::string> m_update_jobs;
   std::set<std::string> m_pgp_import_jobs;
   std::set<std::string> m_cms_import_jobs;
+  std::vector<GpgME::Configuration::Component> m_cached_config;
+  bool m_use_tofu;
 };
 
 KeyCache::KeyCache():
@@ -1142,6 +1324,14 @@ KeyCache::getEncryptionKeys (const std::vector<std::string> &recipients, GpgME::
   return d->getEncryptionKeys (recipients, proto);
 }
 
+std::vector<GpgME::Key>
+KeyCache::getEncryptionKeys (const std::string &recipient, GpgME::Protocol proto) const
+{
+  std::vector<std::string> vec;
+  vec.push_back (recipient);
+  return d->getEncryptionKeys (vec, proto);
+}
+
 static GpgME::Key
 get_most_valid_key_simple (const std::vector<GpgME::Key> &keys)
 {
@@ -1518,7 +1708,7 @@ KeyCache::isMailResolvable(Mail *mail)
   TSTART;
   /* Get the data from the mail. */
   const auto sender = mail->getSender ();
-  auto recps = mail->getCachedRecipients ();
+  auto recps = mail->getCachedRecipientAddresses ();
 
   if (sender.empty() || recps.empty())
     {
@@ -1682,3 +1872,105 @@ KeyCache::import_pgp_key_data (const GpgME::Data &data)
     }
   TRETURN !result.error();
 }
+
+const std::vector<GpgME::Configuration::Component>
+KeyCache::get_cached_config () const
+{
+  return d->get_cached_config ();
+}
+
+void
+KeyCache::setConfig (const std::vector<GpgME::Configuration::Component>& conf)
+{
+  d->setConfig (conf);
+  for (const auto &component: conf)
+   {
+     if (component.name () && !strcmp (component.name (), "gpg"))
+       {
+         for (const auto &option: component.options ())
+           {
+             if (option.name () && !strcmp (option.name (), "trust-model"))
+               {
+                 const char *val = option.currentValue().stringValue();
+                 if (!val)
+                   {
+                     return;
+                   }
+                 if (!strcmp ("tofu", val) || !strcmp ("tofu+pgp", val))
+                   {
+                     log_dbg ("Keycache detected tofu mode.");
+                     d->m_use_tofu = true;
+                     return;
+                   }
+               }
+           }
+         return;
+       }
+   }
+}
+
+bool
+KeyCache::useTofu () const
+{
+  return d->m_use_tofu;
+}
+
+bool
+KeyCache::protocolIsOnline (GpgME::Protocol proto) const
+{
+  TSTART;
+  if (proto == GpgME::OpenPGP && opt.autoretrieve)
+    {
+      TRETURN true;
+    }
+  for (const auto &component: d->get_cached_config ())
+    {
+      if (proto == GpgME::OpenPGP && component.name () && !strcmp (component.name (), "gpg"))
+        {
+          for (const auto &option: component.options ())
+            {
+              if (option.name () && !strcmp (option.name (), "auto-key-retrieve") &&
+                  option.currentValue().boolValue ())
+                {
+                  log_debug ("%s:%s: Detected auto-key-retrieve -> online",
+                             SRCNAME, __func__);
+                  TRETURN true;
+                }
+              if (option.name () && !strcmp (option.name (), "disable-dirmngr") &&
+                  option.currentValue().boolValue ())
+                {
+                  log_debug ("%s:%s: Detected disable-dirmngr -> offline",
+                             SRCNAME, __func__);
+                  TRETURN false;
+                }
+            }
+        }
+      if (proto == GpgME::CMS && component.name () && !strcmp (component.name (), "gpgsm"))
+        {
+          for (const auto &option: component.options ())
+            {
+              if (option.name () && !strcmp (option.name (), "disable-crl-checks") &&
+                  option.currentValue().boolValue ())
+                {
+                  log_debug ("%s:%s: Detected disable-crl-checks -> offline",
+                             SRCNAME, __func__);
+                  TRETURN false;
+                }
+              if (option.name () && !strcmp (option.name (), "disable-dirmngr") &&
+                  option.currentValue().boolValue ())
+                {
+                  log_debug ("%s:%s: Detected disable-dirmngr -> offline",
+                             SRCNAME, __func__);
+                  TRETURN false;
+                }
+            }
+          /* By default S/MIME is online. */
+          log_dbg ("No options found. So assume CRL checks -> online");
+          TRETURN true;
+        }
+    }
+
+  log_dbg ("Detected no online options.");
+
+  TRETURN false;
+}
